/*
  Copyright (c) Members of the EGEE Collaboration. 2004. 
  See http://www.eu-egee.org/partners/ for details on the copyright
  holders.  
  
  Licensed under the Apache License, Version 2.0 (the "License"); 
  you may not use this file except in compliance with the License. 
  You may obtain a copy of the License at 
  
      http://www.apache.org/licenses/LICENSE-2.0 
  
  Unless required by applicable law or agreed to in writing, software 
  distributed under the License is distributed on an "AS IS" BASIS, 
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
  See the License for the specific language governing permissions and 
  limitations under the License.
*/

/*
        glite-ce-job-list: a command line interface to get the list of jobs held by the CREAM service

        Authors: Alvise Dorigo <alvise.dorigo@pd.infn.it>
        Version info: $Id: glite-ce-job-list.cpp,v 1.61.4.1.4.4.2.1.2.3 2011/01/31 14:38:43 adorigo Exp $
*/

/**
 *
 * System Headers
 *
 */
#include <sys/types.h>
#include <algorithm>
#include <unistd.h>
#include <iostream>
#include <getopt.h>
#include <istream>
#include <fstream>
#include <stdio.h>


/**
 *
 * Cream Client API Headers
 *
 */
#include "glite/ce/cream-client-api-c/ConfigurationManager.h"
#include "glite/ce/cream-client-api-c/CreamProxyFactory.h"
#include "glite/ce/cream-client-api-c/CreamProxy_Impl.h"
#include "glite/ce/cream-client-api-c/creamApiLogger.h"
#include "glite/ce/cream-client-api-c/JobIdWrapper.h"
#include "glite/ce/cream-client-api-c/VOMSWrapper.h"
#include "glite/ce/cream-client-api-c/certUtil.h"
#include "glite/ce/cream-client-api-c/file_ex.h"
#include "glite/ce/cream-client-api-c/CEUrl.h"

/**
 *
 * CLI Headers
 *
 */
#include "util/cliUtils.h"
#include "hmessages.h"

/**
 *
 * BOOST Headers
 *
 */
#include "boost/algorithm/string.hpp"
#include "boost/regex.hpp"

using namespace std;
using namespace glite::ce::cream_client_api::soap_proxy;
using namespace glite::ce::cream_client_api::util;
using namespace glite::ce::cream_client_api::cream_exceptions;

Namespace namespaces[] = {};

void printhelp() ;

int main(int argc, char *argv[]) {

  string endpoint;
  bool debug = false;
  char c;
  int option_index = 0;
  string logfile;
  bool redir_out = false;
  bool nomsg=false;
  string outfile;
  bool jobid_on_file = false;
  bool noint = false;
  bool WRITE_LOG_ON_FILE = true;
  string user_conf_file = "";
  creamApiLogger* logger_instance = creamApiLogger::instance();
  log4cpp::Category* log_dev = logger_instance->getLogger();
  string proxyfile, certfile;
  bool specified_proxyfile = false;

  int soap_timeout = 30;

  // Initializes the logger
  log_dev->setPriority( log4cpp::Priority::NOTICE );
  logger_instance->setLogfileEnabled( WRITE_LOG_ON_FILE );

  bool verify_ac_sign = true;

  while( 1 ) {
    static struct option long_options[] = {
      {"help", 0, 0, 'h'},
      {"debug", 0, 0, 'd'},
      {"version", 0, 0, 'v'},
      {"nomsg", 0, 0, 'n'},
      {"logfile", 1, 0, 'l'},
      {"output", 1, 0, 'o'},
      {"noint", 0, 0, 'N'},
      {"conf", 1, 0, 'c'},
      {"proxyfile", 1, 0, 'p'},
      {"timeout", 1, 0, 't'},
      {"donot-verify-ac-sign", 0, 0, 'A'},
      {0, 0, 0, 0}
    };
    c = getopt_long(argc, argv, "nhdvAr:o:l:c:p:t:", long_options, &option_index);
    if ( c == -1 )
      break;

    switch(c) {
    case 'A':
      verify_ac_sign = false;
      break;
    case 'p':
      proxyfile = string(optarg);
      specified_proxyfile = true;
      break;
    case 't':
      soap_timeout = atoi(optarg);
      break;
    case 'N':
      noint = true;
      break;
    case 'v':
      help_messages::printver();
      exit(0);
      break;
    case 'n':
        nomsg=true; // FIXME: remove??
        log_dev->setPriority( log4cpp::Priority::ERROR );
        break;
    case 'h':
      printhelp();
      exit(0);
      break;
    case 'd':
      debug=true;
      log_dev->setPriority( log4cpp::Priority::INFO );
      break;
    case 'o':
      jobid_on_file = true;
      outfile = string(optarg);
      break;
    case 'l':
      redir_out = true;
      logfile = string(optarg);
      break;
    case 'c':
      user_conf_file = string(optarg);
      break;
    default:
        cerr << "Type " << argv[0] << " -h for help" << endl;
        exit(1);
    }
  }
  if(debug && nomsg) {
      log_dev->error("Cannot specify both --debug and --nomsg options. Stop.");
      exit(1);
  }

//  CreamProxy *initSoap;
  string VO = "";
  try {
  //  initSoap = new CreamProxy(false);
    if(!specified_proxyfile)
      certfile = cliUtils::getProxyCertificateFile();
    else
      certfile = proxyfile;
    if ( log_dev->isInfoEnabled() ) {
        log_dev->info("Using certificate proxy file [%s]", certfile.c_str() );
    }
//    VO = initSoap->Authenticate(certfile);

    VOMSWrapper V( certfile, verify_ac_sign );
    if( !V.IsValid( ) ) {
      
      logger_instance->log(log4cpp::Priority::FATAL, 
			   string("Problems with proxyfile [")+certfile+"]: "+
			   V.getErrorMessage(), true, true, true);
      exit(1);
    }
//     } else
// 	logger_instance->log(log4cpp::Priority::ERROR, 
// 			     "Error while reading proxyfile ["+certfile+"]: "+
// 			     V.getErrorMessage(), true, true, true);
//       exit(1);
//     }
    
//     long int leftcert = V.getProxyTimeEnd( ) - time( NULL );
    
//     if(leftcert <= 0) {
//       log_dev->fatal( "Proxy certificate is expired. Please Check it." );
//       exit(1);
//     }

    
    
    VO = V.getVOName( );

  } catch(soap_ex& s) {
      log_dev->error( s.what() );
      exit(1);
  } catch(auth_ex& auth) {
      log_dev->error( auth.what() );
      exit(1);
  }


//   if( !cliUtils::checkEndpointFormat( endpoint ) )
//   {
//     log_dev->error( "Endpoint not specified in the right format: should be <host>[:tcpport]. Stop.");
//     exit(1);
//   }


  /**
   * Handles the configuration loading
   */
  vector<string> confiles;
  try {
    confiles = cliUtils::getConfigurationFiles(VO,
					       user_conf_file,
					       DEFAULT_CONF_FILE);
  } catch(no_user_confile_ex& ex) {
      log_dev->error( ex.what() );
    exit(1);
  }

  if(log_dev->isInfoEnabled()) {
      for(vector<string>::const_iterator fit = confiles.begin();
          fit != confiles.end();
	  fit++) { //unsigned int j=0; j<confiles.size(); j++) {
          log_dev->info("Loading configurations from [%s]", (*fit).c_str() );
      }
      log_dev->info("**************************************************************************");
      log_dev->info( help_messages::getStartMessage() );
  }

  ConfigurationManager confMgr(ConfigurationManager::classad);
  try { confMgr.load(confiles); }
  catch(file_ex& ex) {
    if(debug)
      log_dev->warn( string(ex.what())+". Using built-in configuration" );
  }

  if(debug || redir_out) {
      if(!redir_out) {
          logfile = logger_instance->getLogFileName(confMgr.getProperty("LIST_LOG_DIR", DEFAULT_LOG_DIR).c_str(), "glite-ce-job-list");
      }
      log_dev->info( "Logfile is [%s]", logfile.c_str() );
      cliUtils::mkdir(cliUtils::getPath(logfile));
      logger_instance->setLogFile(logfile.c_str());
  }

  if(argv[optind]==NULL) {
      log_dev->error( "Endpoint not specified in the command line arguments. Stop.");
      exit(1);
  } else {
    endpoint = string(argv[optind]);

    if( !cliUtils::checkEndpointFormat( endpoint ) )
    {
      log_dev->error( "Endpoint not specified in the right format: should be <host>[:tcpport]. Stop.");
      exit(1);
    }

  }


//   if( !cliUtils::checkEndpointFormat( endpoint ) )
//   {
//     log_dev->error( "Endpoint not specified in the right format: should be <host>[:tcpport]. Stop.");
//     exit(1);
//   }


  /**
   *
   */

//   boost::regex tmp2( "^\\w+(:\\d+)" );
//     boost::smatch what2;
//     if( !boost::regex_match( endpoint, what2, tmp2 ) ) {
  if(!cliUtils::containsTCPPort( endpoint )) {
    //if( what2[1] == "" )
    endpoint = endpoint + ":" + confMgr.getProperty("DEFAULT_CREAM_TCPPORT", "8443");
  }

  string serviceAddress = confMgr.getProperty("CREAM_URL_PREFIX", DEFAULT_PROTO)+
    endpoint+"/"+confMgr.getProperty("CREAM_URL_POSTFIX",DEFAULT_CEURL_POSTFIX);//CEUrl::getCreamURI(URL);

  if (log_dev->isInfoEnabled()) {
      log_dev->info( "Service address=[%s]", serviceAddress.c_str() );
  }
  //printf("Service URL: %s\n",  serviceAddress.c_str());
  vector<JobIdWrapper> jobIdWrapperVetor;
  
  AbsCreamProxy* creamClient;
  creamClient = CreamProxyFactory::make_CreamProxyList(&jobIdWrapperVetor, soap_timeout );
  if(!creamClient)
    {
      logger_instance->log(log4cpp::Priority::FATAL, "FAILED CREATION OF AN AbsCreamProxy object! STOP!!", true, true, true);
      return 1;
    }

  creamClient->setCredential( certfile );
  try {

     creamClient->execute(serviceAddress);
     //creamClient->garbage_collect();

   } catch(BaseException& ex ) {
      log_dev->log(log4cpp::Priority::ERROR, ex.what(), true, WRITE_LOG_ON_FILE, true);
      exit(1);
   } catch (exception& ex) {

     log_dev->error(ex.what());
     return 1;
   }

   string prefixJobId = confMgr.getProperty("CREAM_URL_PREFIX", DEFAULT_PROTO)+ endpoint + "/";
// try {
//    initSoap->List(serviceAddress.c_str(), result);
//  } catch(exception& ex) {
//      log_dev->log(log4cpp::Priority::ERROR, ex.what(), true, WRITE_LOG_ON_FILE, true);
//      exit(1);
//  }
  /*catch(InternalException& ex) {
    log_dev->log(log4cpp::Priority::ERROR, ex.what(), true, WRITE_LOG_ON_FILE, true);
    exit(1);
  }*/

  //logger::instance().log(logger::INFO, "Info succesfull", true, true, true);
  if(log_dev->isInfoEnabled()) {
     log_dev->info( "Logfile is [%s]", logfile.c_str() );
   }

  //int f = -1;
  if(jobid_on_file) {
     bool exists = false;

     if((exists=cliUtils::fileExists(outfile.c_str()))==true) {
	if(!cliUtils::fileIsWritable(outfile.c_str())){
            log_dev->error("Output file ["+outfile+"] exists but is not accessible. Stop" );
	  exit(1);
	}
     }
     // OK, the output file does exist or doesn't
     // if it does let's check if the first line has the magic string
     int f;
     if(exists) {
        bool isjoblist = cliUtils::isACreamJobListFile(outfile.c_str());
	if(!isjoblist) {
	 if(!noint) {
	   printf("\nWARNING: file [%s] already exists and is not a CREAM job list file. Overwrite it (y/n) ? ", outfile.c_str());
	   char c; cin >> c;
	   if(c!='y') {
	     printf("\nListing aborted. Bye!\n"); exit(0);
	   }
         }
	 // Is NOT a JobList => opens file and truncs it
	 try { f=cliUtils::openJobListFile(outfile.c_str(), true); }
	 catch(file_ex& fex) {
             log_dev->error( fex.what() );
             exit(1);
	 }
        } else {  // if(!isjoblist)
	  // opens file in append mode
	  try { f=cliUtils::appendToJobListFile(outfile.c_str()); }
	  catch(file_ex& fex) {
              log_dev->error( fex.what() );
              exit(1);
	  }
	}
     } // if(exists)
     else { //if(exist)
     	// create file from scratch
	try { f=cliUtils::openJobListFile(outfile.c_str(), false); }
	catch(file_ex& fex) {
	    log_dev->error( fex.what() );
	    exit(1);
	}
     }
     // Writes JobIDs into file; but before doing that
     // retrieves the list of jobs already in the file to prevent
     // duplications
     vector<string> alreadyThere;
     alreadyThere.reserve(1000);
     cliUtils::getJobIDFromFile(alreadyThere, outfile.c_str());
     int count = 0;
     for(vector<JobIdWrapper>::const_iterator rit = jobIdWrapperVetor.begin();
         rit != jobIdWrapperVetor.end();
	 rit++) { //unsigned int j=0; j<result.size(); j++) {
       //if(string_manipulation::vector_contains( alreadyThere, result[j]) ) {
       string creamJobId = (*rit).getCreamURL();
       string::size_type loc = creamJobId.find(confMgr.getProperty("CREAM_URL_POSTFIX", DEFAULT_CEURL_POSTFIX));
       if( loc != string::npos) {
         creamJobId = creamJobId.substr(0,loc);
        }
       creamJobId += "/";
       creamJobId.append((*rit).getCreamJobID());
       if( find(alreadyThere.begin(), alreadyThere.end(), creamJobId ) != alreadyThere.end()){
           log_dev->warn( "JobID [" + creamJobId + "] already in the list. Skipping it..." );
           continue;
        }
	try { cliUtils::writeJobID(f, creamJobId ); }
	catch(file_ex& fex) {
            log_dev->error( fex.what() );
	  close(f);
	  exit(1);
	}
	count++;
     }
     if(log_dev->isInfoEnabled()) {
       log_dev->info( "Wrote %d job(s) into file [%s]", count, outfile.c_str() );
     }
  } else {
    for(vector<JobIdWrapper>::const_iterator rit = jobIdWrapperVetor.begin();
        rit != jobIdWrapperVetor.end();
	rit++) { //unsigned int j=0; j<result.size(); j++) {
        string creamJobId = (*rit).getCreamURL();
        string::size_type loc = creamJobId.find(confMgr.getProperty("CREAM_URL_POSTFIX", DEFAULT_CEURL_POSTFIX));
        if( loc != string::npos) {
          creamJobId = creamJobId.substr(0,loc);
        }
	creamJobId += "/";
        creamJobId.append((*rit).getCreamJobID());  
        cout << creamJobId.c_str() << endl;
      //log_dev->log( log4cpp::Priority::INFO, result[j]);
    }
   }
  //delete(initSoap);

   //return err;

  return 0;
}

void printhelp() {
  printf("%s\n",help_messages::getListHelp().c_str());
}
